#ifndef SIPM_SIPM_H
#define SIPM_SIPM_H

#define SIPM_VERSION "2.3.1"

#include "SiPMAnalogSignal.h"
#include "SiPMDebugInfo.h"
#include "SiPMHit.h"
#include "SiPMProperties.h"
#include "SiPMRandom.h"
#include "SiPMSensor.h"
#include "SiPMTypes.h"

#endif
